# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import pickle
from math import log
from time import time

import linien_gui
import numpy as np
from linien_common.common import check_plot_data
from linien_common.config import N_COLORS
from linien_gui.config import COLORS
from linien_gui.ui.plot_widget import INVALID_POWER
from linien_gui.utils_gui import color_to_hex
from linien_gui.widgets import CustomWidget
from PyQt5 import QtWidgets

ZOOM_STEP = 0.9
MAX_ZOOM = 50
MIN_ZOOM = 0


def sweep_amplitude_to_zoom_step(amplitude):
    return round(log(amplitude, ZOOM_STEP))


class MainWindow(QtWidgets.QMainWindow, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui("main_window.ui")

        self.reset_std_history()

    def show(self, host, name):
        self.setWindowTitle(
            "Linien spectroscopy lock %s: %s (%s)"
            % (linien_gui.__version__, name, host)
        )
        super().show()

    def closeEvent(self, *args, **kwargs):
        self.app.close_all_secondary_windows()
        super().closeEvent(*args, **kwargs)

    def ready(self):
        # handle keyboard events
        self.setFocus()

        self.ids.export_parameters_button.clicked.connect(
            self.export_parameters_select_file
        )
        self.ids.import_parameters_button.clicked.connect(self.import_parameters)

        def display_power(power, element):
            if power == INVALID_POWER:
                element.hide()
            else:
                element.show()
                text = "%.2f" % power
                element.setText(text)

        def display_power_channel_1(power):
            el = self.ids.power_channel_1
            display_power(power, el)

        def display_power_channel_2(power):
            el = self.ids.power_channel_2
            display_power(power, el)

        self.ids.graphicsView.signal_power1.connect(display_power_channel_1)
        self.ids.graphicsView.signal_power2.connect(display_power_channel_2)
        self.ids.graphicsView.keyPressed.connect(self.handle_key_press)

        # by default we hide it and just show when a new version is available
        self.ids.new_version_available_label.hide()

    def show_new_version_available(self):
        self.ids.new_version_available_label.show()

    def handle_key_press(self, key):
        print("key pressed", key)

    def export_parameters_select_file(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".json"
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            "JSON (*%s)" % default_ext,
            options=options,
        )
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext

            with open(fn, "w") as f:
                json.dump(
                    {
                        "linien-version": linien_gui.__version__,
                        "time": time(),
                        "parameters": dict(
                            (k, getattr(self.parameters, k).value)
                            for k in self.parameters.remote.exposed_get_restorable_parameters()  # noqa: E501
                        ),
                    },
                    f,
                )

    def import_parameters(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".json"
        fn, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            "JSON (*%s)" % default_ext,
            options=options,
        )
        if fn:
            with open(fn, "r") as f:
                data = json.load(f)

            assert "linien-version" in data, "invalid parameter file"

            restorable = self.parameters.remote.exposed_get_restorable_parameters()
            for k, v in data["parameters"].items():
                if k not in restorable:
                    print("ignore key", k)
                    continue

                print("restoring", k)
                getattr(self.parameters, k).value = v

            self.control.write_registers()

    def connection_established(self):
        self.control = self.app.control
        self.parameters = self.app.parameters

        def change_sweep_control_visibility(*args):
            al_running = self.parameters.autolock_running.value
            optimization = self.parameters.optimization_running.value
            locked = self.parameters.lock.value

            self.get_widget("sweep_control").setVisible(
                not al_running and not locked and not optimization
            )
            self.get_widget("top_lock_panel").setVisible(locked)
            self.get_widget("statusbar_unlocked").setVisible(
                not al_running and not locked and not optimization
            )

        self.parameters.lock.on_change(change_sweep_control_visibility)
        self.parameters.autolock_running.on_change(change_sweep_control_visibility)
        self.parameters.optimization_running.on_change(change_sweep_control_visibility)

        self.parameters.to_plot.on_change(self.update_std)

        self.parameters.pid_on_slow_enabled.on_change(
            lambda v: self.ids.legend_slow_signal_history.setVisible(v)
        )
        self.parameters.dual_channel.on_change(
            lambda v: self.ids.legend_monitor_signal_history.setVisible(not v)
        )

        self.ids.settings_toolbox.setCurrentIndex(0)

        self.parameters.lock.on_change(lambda *args: self.reset_std_history())

        def update_legend_color(*args):
            def set_color(el, color_name):
                return el.setStyleSheet(
                    "color: "
                    + color_to_hex(
                        getattr(
                            self.parameters, "plot_color_%d" % COLORS[color_name]
                        ).value
                    )
                )

            set_color(self.ids.legend_spectrum_1, "spectrum_1")
            set_color(self.ids.legend_spectrum_2, "spectrum_2")
            set_color(self.ids.legend_spectrum_combined, "spectrum_combined")
            set_color(self.ids.legend_error_signal, "spectrum_combined")
            set_color(self.ids.legend_control_signal, "control_signal")
            set_color(self.ids.legend_control_signal_history, "control_signal_history")
            set_color(self.ids.legend_slow_signal_history, "slow_history")
            set_color(self.ids.legend_monitor_signal_history, "monitor_signal_history")

        for color_idx in range(N_COLORS):
            getattr(self.parameters, "plot_color_%d" % color_idx).on_change(
                update_legend_color
            )

        def update_legend_text(dual_channel):
            self.ids.legend_spectrum_1.setText(
                "error signal" if not dual_channel else "error signal 1"
            )
            self.ids.legend_spectrum_2.setText(
                "monitor" if not dual_channel else "error signal 2"
            )
            self.ids.legend_spectrum_combined.setVisible(dual_channel)

        self.parameters.dual_channel.on_change(update_legend_text)

    def update_std(self, to_plot, max_std_history_length=10):
        if self.parameters.lock.value and to_plot:
            to_plot = pickle.loads(to_plot)
            if to_plot and check_plot_data(True, to_plot):
                error_signal = to_plot.get("error_signal")
                control_signal = to_plot.get("control_signal")

                self.error_std_history.append(np.std(error_signal))
                self.control_std_history.append(np.std(control_signal))

                self.error_std_history = self.error_std_history[
                    -max_std_history_length:
                ]
                self.control_std_history = self.control_std_history[
                    -max_std_history_length:
                ]

                if error_signal is not None and control_signal is not None:
                    self.ids.error_std.setText("%.2f" % np.mean(self.error_std_history))
                    self.ids.control_std.setText(
                        "%.2f" % np.mean(self.control_std_history)
                    )

    def reset_std_history(self):
        self.error_std_history = []
        self.control_std_history = []



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import linien_gui
from linien_gui.config import load_device_data, save_device_data
from linien_gui.dialogs import (
    LoadingDialog,
    ask_for_parameter_restore_dialog,
    error_dialog,
    question_dialog,
    show_installation_progress_widget,
)
from linien_gui.threads import ConnectionThread
from linien_gui.ui.new_device_dialog import NewDeviceDialog
from linien_gui.utils_gui import set_window_icon
from linien_gui.widgets import CustomWidget
from PyQt5 import QtCore, QtWidgets


class DeviceManager(QtWidgets.QMainWindow, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui("device_manager.ui")
        self.setWindowTitle(f"Linien spectroscopy lock v{linien_gui.__version__}")
        set_window_icon(self)

    def ready(self):
        self.load_device_data(autoload=True)

    def keyPressEvent(self, event):
        key = event.key()

        if key in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            self.connect()

    def load_device_data(self, autoload=False):
        devices = load_device_data()
        lst = self.ids.deviceList
        lst.clear()

        for device in devices:
            lst.addItem("{} ({})".format(device["name"], device["host"]))

        if autoload and len(devices) == 1:
            self.connect_to_device(devices[0])

    def connect(self):
        devices = load_device_data()

        if not devices:
            return
        else:
            self.connect_to_device(devices[self.get_list_index()])

    def connect_to_device(self, device):
        loading_dialog = LoadingDialog(self, device["host"])
        loading_dialog.show()

        aborted = {}

        self.connection_thread = ConnectionThread(device)

        def was_aborted(*args):
            aborted["aborted"] = True

        loading_dialog.aborted.connect(was_aborted)

        # Define slot functions to be connected ----------------------------------------
        def on_client_connected(client):
            loading_dialog.hide()
            if not aborted:
                self.app.client_connected(client)

        def handle_server_not_installed():
            loading_dialog.hide()
            if not aborted:
                display_question = (
                    "The server is not yet installed on the device. Should it be "
                    "installed? (Requires internet connection on RedPitaya)"
                )
                if question_dialog(self, display_question, "Install server?"):
                    show_installation_progress_widget(
                        parent=self,
                        device=device,
                        callback=lambda: self.connect_to_device(device),
                    )

        def handle_invalid_server_version(
            remote_version: str, client_version: str
        ) -> None:
            loading_dialog.hide()
            if not aborted:
                display_question = (
                    f"Server version ({remote_version}) does not match the client "
                    f"({client_version}) version. Should the corresponding server "
                    f"version be installed?"
                )
                if question_dialog(
                    self, display_question, "Install corresponding version?"
                ):
                    show_installation_progress_widget(
                        parent=self,
                        device=device,
                        callback=lambda: self.connect_to_device(device),
                    )

        def handle_authentication_exception():
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    "Error at authentication.\n"
                    "Check username and password (by default both are 'root')  and "
                    "verify that you don't have any offending SSH keys in your known "
                    "hosts file."
                )
                error_dialog(self, display_error)

        def handle_general_connection_error():
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    "Unable to connect to device. If you are connecting by hostname "
                    "(i.e. `rp-xxxxxx.local`), try using IP address instead."
                )
                error_dialog(self, display_error)

        def handle_other_exception(exception):
            loading_dialog.hide()
            if not aborted:
                display_error = (
                    f"Exception occured when connecting to the device:\n\n {exception}"
                )
                error_dialog(self, display_error)

        def ask_for_parameter_restore():
            question = (
                "Linien on RedPitaya is running with different parameters than the "
                "ones saved locally on this machine. Do you want to upload the local "
                "parameters or keep the remote ones? Note that remote parameters are "
                "only saved if Linien server was shut down properly, not when "
                "unplugging the power plug. In this case, you should update your local "
                "parameters."
            )
            should_restore = ask_for_parameter_restore_dialog(
                self, question, "Restore parameters?"
            )
            self.connection_thread.answer_whether_to_restore_parameters(should_restore)

        def handle_connection_lost():
            error_dialog(self, "Lost connection to the server!")
            self.app.close()

        # Connect slots to signals -----------------------------------------------------
        self.connection_thread.client_connected.connect(on_client_connected)

        self.connection_thread.server_not_installed_exception_raised.connect(
            handle_server_not_installed
        )
        self.connection_thread.invalid_server_version_exception_raised.connect(
            handle_invalid_server_version
        )
        self.connection_thread.authentication_exception_raised.connect(
            handle_authentication_exception
        )
        self.connection_thread.general_connection_exception_raised.connect(
            handle_general_connection_error
        )
        self.connection_thread.other_exception_raised.connect(handle_other_exception)
        self.connection_thread.ask_for_parameter_restore.connect(
            ask_for_parameter_restore
        )
        self.connection_thread.connection_lost.connect(handle_connection_lost)

        # Start the worker -------------------------------------------------------------
        self.connection_thread.start()

    def new_device(self):
        self.dialog = NewDeviceDialog()
        self.dialog.setModal(True)
        self.dialog.show()

        def reload_device_data():
            # not very elegant...
            QtCore.QTimer.singleShot(100, self.load_device_data)

        self.dialog.accepted.connect(reload_device_data)

    def edit_device(self):
        devices = load_device_data()

        if not devices:
            return

        device = devices[self.get_list_index()]

        self.dialog = NewDeviceDialog(device)
        self.dialog.setModal(True)
        self.dialog.show()

        def reload_device_data():
            # not very elegant...
            QtCore.QTimer.singleShot(100, self.load_device_data)

        self.dialog.accepted.connect(reload_device_data)

    def move_device_up(self):
        self.move_device(-1)

    def move_device_down(self):
        self.move_device(1)

    def move_device(self, direction):
        devices = load_device_data()

        if not devices:
            return

        current_index = self.get_list_index()
        new_index = current_index + direction

        if new_index < 0 or new_index > len(devices) - 1:
            return

        device = devices.pop(current_index)
        devices = devices[:new_index] + [device] + devices[new_index:]
        save_device_data(devices)
        self.load_device_data()
        self.ids.deviceList.setCurrentRow(new_index)

    def get_list_index(self):
        return self.ids.deviceList.currentIndex().row()

    def remove_device(self):
        devices = load_device_data()

        if not devices:
            return

        devices.pop(self.get_list_index())
        save_device_data(devices)
        self.load_device_data()

    def selected_device_changed(self):
        idx = self.get_list_index()

        disable_buttons = True

        if idx >= 0:
            devices = load_device_data()

            if devices:
                disable_buttons = False

        for btn in [
            self.ids.connectButton,
            self.ids.removeButton,
            self.ids.editButton,
            self.ids.moveUpButton,
            self.ids.moveDownButton,
        ]:
            btn.setEnabled(not disable_buttons)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import HIGH_PASS_FILTER, LOW_PASS_FILTER
from linien_gui.utils_gui import param2ui
from linien_gui.widgets import CustomWidget
from PyQt5 import QtWidgets


class SpectroscopyPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args):
        super().__init__(*args)
        self.load_ui("spectroscopy_panel.ui")

        def change_filter_frequency(filter_i):
            self.get_param("filter_%d_frequency" % filter_i).value = getattr(
                self.ids, "filter_%d_frequency" % filter_i
            ).value()
            self.control.write_registers()

        def change_filter_enabled(filter_i):
            filter_enabled = int(
                getattr(self.ids, "filter_%d_enabled" % filter_i).checkState() > 0
            )
            self.get_param("filter_%d_enabled" % filter_i).value = filter_enabled
            self.control.write_registers()

        def change_filter_type(filter_i):
            param = self.get_param("filter_%d_type" % filter_i)
            current_idx = getattr(self.ids, "filter_%d_type" % filter_i).currentIndex()
            param.value = (LOW_PASS_FILTER, HIGH_PASS_FILTER)[current_idx]
            self.control.write_registers()

        for filter_i in [1, 2]:
            for key, fct in {
                "change_filter_%d_frequency": change_filter_frequency,
                "change_filter_%d_enabled": change_filter_enabled,
                "change_filter_%d_type": change_filter_type,
            }.items():
                setattr(
                    self,
                    key % filter_i,
                    lambda *args, fct=fct, filter_i=filter_i: fct(filter_i),
                )

    def get_param(self, name):
        return getattr(
            self.parameters, "%s_%s" % (name, "a" if self.channel == 0 else "b")
        )

    def ready(self):
        self.ids.signal_offset.setKeyboardTracking(False)
        self.ids.signal_offset.valueChanged.connect(self.change_signal_offset)
        self.ids.demodulation_phase.setKeyboardTracking(False)
        self.ids.demodulation_phase.valueChanged.connect(self.change_demod_phase)
        self.ids.demodulation_frequency.currentIndexChanged.connect(
            self.change_demod_multiplier
        )

        for filter_i in [1, 2]:
            _get = lambda parent, attr, filter_i=filter_i: getattr(
                parent, attr % filter_i
            )
            _get(self.ids, "filter_%d_enabled").stateChanged.connect(
                _get(self, "change_filter_%d_enabled")
            )
            freq_input = _get(self.ids, "filter_%d_frequency")
            freq_input.setKeyboardTracking(False)
            freq_input.valueChanged.connect(_get(self, "change_filter_%d_frequency"))
            _get(self.ids, "filter_%d_type").currentIndexChanged.connect(
                _get(self, "change_filter_%d_type")
            )

        def automatic_changed(value):
            self.get_param("filter_automatic").value = value
            self.control.write_registers()

        self.ids.filter_automatic.stateChanged.connect(automatic_changed)

        def invert_changed(value):
            self.get_param("invert").value = bool(value)
            self.control.write_registers()

        self.ids.invert.stateChanged.connect(invert_changed)

    def connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        # self.close_button.clicked.connect(self.close_app)
        # self.shutdown_button.clicked.connect(self.shutdown_server)

        param2ui(self.get_param("demodulation_phase"), self.ids.demodulation_phase)
        param2ui(
            self.get_param("demodulation_multiplier"),
            self.ids.demodulation_frequency,
            lambda value: value - 1,
        )
        param2ui(self.get_param("offset"), self.ids.signal_offset, lambda v: v / 8191.0)
        param2ui(self.get_param("invert"), self.ids.invert)
        param2ui(self.get_param("filter_automatic"), self.ids.filter_automatic)

        def filter_automatic_changed(value):
            self.ids.automatic_filtering_enabled.setVisible(value)
            self.ids.automatic_filtering_disabled.setVisible(not value)

        self.get_param("filter_automatic").on_change(filter_automatic_changed)

        for filter_i in [1, 2]:
            param2ui(
                self.get_param("filter_%d_enabled" % filter_i),
                getattr(self.ids, "filter_%d_enabled" % filter_i),
            )
            param2ui(
                self.get_param("filter_%d_frequency" % filter_i),
                getattr(self.ids, "filter_%d_frequency" % filter_i),
            )
            param2ui(
                self.get_param("filter_%d_type" % filter_i),
                getattr(self.ids, "filter_%d_type" % filter_i),
                lambda type_: {LOW_PASS_FILTER: 0, HIGH_PASS_FILTER: 1}[type_],
            )

    def change_signal_offset(self):
        self.get_param("offset").value = self.ids.signal_offset.value() * 8191
        self.control.write_registers()

    def change_demod_phase(self):
        self.get_param("demodulation_phase").value = self.ids.demodulation_phase.value()
        self.control.write_registers()

    def change_demod_multiplier(self, idx):
        self.get_param("demodulation_multiplier").value = idx + 1
        self.control.write_registers()


class SpectroscopyChannelAPanel(SpectroscopyPanel):
    channel = 0


class SpectroscopyChannelBPanel(SpectroscopyPanel):
    channel = 1



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import os
import pickle

import appdirs
import rpyc

COLORS = {
    "spectrum_1": 0,
    "spectrum_2": 1,
    "spectrum_combined": 2,
    "control_signal": 0,
    "control_signal_history": 1,
    "slow_history": 3,
    "monitor_signal_history": 4,
}
# don't plot more often than once per `DEFAULT_PLOT_RATE_LIMIT` seconds
DEFAULT_PLOT_RATE_LIMIT = 0.1


def get_data_folder():
    folder_name = appdirs.user_data_dir("linien")

    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    return folder_name


def get_devices_filename():
    return os.path.join(get_data_folder(), "devices")


def save_device_data(devices):
    with open(get_devices_filename(), "wb") as f:
        pickle.dump(devices, f)


def load_device_data():
    try:
        with open(get_devices_filename(), "rb") as f:
            devices = pickle.load(f)
    except (FileNotFoundError, pickle.UnpicklingError, EOFError):
        devices = []

    return devices


def save_parameter(device_key, param, value, delete=False):
    devices = load_device_data()
    device = [d for d in devices if d["key"] == device_key][0]
    device.setdefault("params", {})

    if not delete:
        # FIXME: This is the only part where rpyc is used in linien-gui. Remove it if
        # possible.
        # rpyc obtain is for ensuring that we don't try to save a netref here
        try:
            device["params"][param] = rpyc.classic.obtain(value)
        except Exception:
            print("unable to obtain and save parameter", param)
    else:
        try:
            del device["params"][param]
        except KeyError:
            pass

    save_device_data(devices)


def get_saved_parameters(device_key):
    devices = load_device_data()
    device = [d for d in devices if d["key"] == device_key][0]
    device.setdefault("params", {})
    return device["params"]



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import traceback
from queue import Queue

from linien_client.connection import LinienClient
from linien_client.deploy import install_remote_server
from linien_client.exceptions import (
    GeneralConnectionErrorException,
    InvalidServerVersionException,
    RPYCAuthenticationException,
    ServerNotInstalledException,
)
from linien_common.config import DEFAULT_SERVER_PORT
from linien_gui.config import get_saved_parameters, save_parameter
from PyQt5.QtCore import QThread, pyqtSignal


class FileLikeQueue(Queue):
    """
    A queue with the interface of a file-like object used to redirect output stream of
    SSH calls to the GUI.
    """

    def write(self, data):
        self.put(data)

    def read(self):
        return self.get()

    def flush(self):
        pass


class RemoteServerInstallationThread(QThread):
    def __init__(self, device):
        """A thread that installs the linien server on a remote machine."""
        super().__init__()
        self.device = device

    out_stream = FileLikeQueue()

    def run(self):
        install_remote_server(
            host=self.device["host"],
            user=self.device["username"],
            password=self.device["password"],
            out_stream=self.out_stream,
        )


class ConnectionThread(QThread):
    def __init__(self, device):
        super().__init__()
        self.device = device

    client_connected = pyqtSignal(object)
    server_not_installed_exception_raised = pyqtSignal()
    invalid_server_version_exception_raised = pyqtSignal(str, str)
    authentication_exception_raised = pyqtSignal()
    general_connection_exception_raised = pyqtSignal()
    other_exception_raised = pyqtSignal(str)
    connection_lost = pyqtSignal()
    ask_for_parameter_restore = pyqtSignal()

    def run(self):
        try:
            self.client = LinienClient(
                host=self.device["host"],
                user=self.device["username"],
                password=self.device["password"],
                port=self.device.get("port", DEFAULT_SERVER_PORT),
                name=self.device.get("name", ""),
            )
            self.client.connect(
                autostart_server=True,
                use_parameter_cache=True,
                call_on_error=self.on_connection_lost,
            )
            self.client_connected.emit(self.client)

        except ServerNotInstalledException:
            self.server_not_installed_exception_raised.emit()

        except InvalidServerVersionException as e:
            return self.invalid_server_version_exception_raised.emit(
                e.remote_version, e.client_version
            )

        except RPYCAuthenticationException:
            self.authentication_exception_raised.emit()

        except GeneralConnectionErrorException:
            self.general_connection_exception_raised.emit()

        except Exception:
            traceback.print_exc()
            self.other_exception_raised.emit(traceback.format_exc())

        # now, we are connected to the server. Check whether we have cached settings for
        # this server. If yes, check whether they match with what is currentlyrunning.
        # If there is a mismatch, ask the user whether the settings should be restored.

        parameters_differ = self.restore_parameters(dry_run=True)
        if parameters_differ:
            self.ask_for_parameter_restore.emit()
        else:
            # if parameters don't differ, we can start monitoring remote parameter
            # changes and write them to disk. We don't do this if parameters differ
            # because we don't want to override our local settings with the remote one
            # --> we wait until user has answered whether local parameters or remote
            # ones should be used.
            self.continuously_write_parameters_to_disk()

    def on_connection_lost(self):
        self.connection_lost.emit()

    def answer_whether_to_restore_parameters(self, should_restore):
        if should_restore:
            self.restore_parameters(dry_run=False)

        self.continuously_write_parameters_to_disk()

    def restore_parameters(self, dry_run=False):
        """
        Reads settings for a server that were cached locally. Sends them to the server.
        If `dry_run` is...

            * `True`, this function returns a boolean indicating whether the
              local parameters differ from the ones on the server
            * `False`, the local parameters are uploaded to the server
        """
        device_key = self.device["key"]
        params = get_saved_parameters(device_key)
        print("restoring parameters")

        differences = False

        for k, v in params.items():
            if hasattr(self.client.parameters, k):
                param = getattr(self.client.parameters, k)
                if param.value != v:
                    if dry_run:
                        print("parameter", k, "differs")
                        differences = True
                        break
                    else:
                        param.value = v
            else:
                # This may happen if the settings were written with a different version
                # of linien.
                print(f"Unable to restore parameter {k}. Delete the cached value.")
                save_parameter(device_key, k, None, delete=True)

        if not dry_run:
            self.client.control.write_registers()

        return differences

    def continuously_write_parameters_to_disk(self):
        """
        Listens for changes of some parameters and permanently saves their values on the
        client's disk. This data can be used to restore the status later, if the client
        tries to connect to the server but it doesn't run anymore.
        """
        params = self.client.parameters.remote.exposed_get_restorable_parameters()

        for param in params:

            def on_change(value, param=param):
                save_parameter(self.device["key"], param, value)

            getattr(self.client.parameters, param).on_change(on_change)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import weakref
from pathlib import Path

from PyQt5 import uic
from pyqtgraph.Qt import QtCore

UI_PATH = Path(__file__).parents[0].resolve() / "ui"


class IDSelector:
    def __init__(self, parent):
        self.parent = parent

    def __getattr__(self, name):
        return self.parent.get_widget(name)


class CustomWidget:
    instances = []

    def __init__(self, *args, **kwargs):
        self.__class__.instances.append(weakref.proxy(self))

        super().__init__(*args, **kwargs)

        self.ids = IDSelector(self)

        QtCore.QTimer.singleShot(100, self.ready)

    def ready(self):
        pass

    def connection_established(self):
        # This is executed the client succesfully established a connection to the server
        # and can be extended by inheritting classes.
        pass

    def get_widget(self, name):
        """Queries a widget by name."""
        return self.findChild(QtCore.QObject, name)

    @property
    def app(self):
        # this property is set manually. Probably there is a more elegant way
        # to solve this...
        return self.window()._app

    @app.setter
    def app(self, app):
        self._app = app

    def load_ui(self, name):
        assert name.endswith(".ui")
        path = UI_PATH / name
        uic.loadUi(str(path), self)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import (
    ANALOG_OUT0,
    ANALOG_OUT_V,
    FAST_OUT1,
    FAST_OUT2,
    convert_channel_mixing_value,
)
from linien_gui.utils_gui import param2ui
from linien_gui.widgets import CustomWidget
from PyQt5 import QtWidgets


class GeneralPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui("general_panel.ui")

    def ready(self):
        self.ids.channel_mixing_slider.valueChanged.connect(self.channel_mixing_changed)
        self.ids.fast_mode.stateChanged.connect(self.fast_mode_changed)
        self.ids.dual_channel.stateChanged.connect(self.dual_channel_changed)

        self.ids.mod_channel.currentIndexChanged.connect(self.mod_channel_changed)
        self.ids.control_channel.currentIndexChanged.connect(
            self.control_channel_changed
        )
        self.ids.sweep_channel.currentIndexChanged.connect(self.sweep_channel_changed)
        self.ids.slow_control_channel.currentIndexChanged.connect(
            self.slow_control_channel_changed
        )

        self.ids.polarity_fast_out1.currentIndexChanged.connect(
            self.polarity_fast_out1_changed
        )
        self.ids.polarity_fast_out2.currentIndexChanged.connect(
            self.polarity_fast_out2_changed
        )
        self.ids.polarity_analog_out0.currentIndexChanged.connect(
            self.polarity_analog_out0_changed
        )

        for idx in range(4):
            if idx == 0:
                continue
            element = getattr(self.ids, "analog_out_%d" % idx)
            element.setKeyboardTracking(False)
            element.valueChanged.connect(
                lambda value, idx=idx: self.change_analog_out(idx)
            )

    def change_analog_out(self, idx):
        name = "analog_out_%d" % idx
        getattr(self.parameters, name).value = int(
            getattr(self.ids, name).value() / ANALOG_OUT_V
        )
        self.control.write_registers()

    def connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.parameters.fast_mode, self.ids.fast_mode)

        param2ui(self.parameters.fast_mode, self.ids.fast_mode)

        def dual_channel_changed(value):
            self.ids.dual_channel_mixing.setVisible(value)
            self.ids.fast_in_1_status.setText(
                "error signal" if not value else "error signal 1"
            )
            self.ids.fast_in_2_status.setText(
                "monitor" if not value else "error signal 2"
            )
            return value

        param2ui(
            self.parameters.dual_channel, self.ids.dual_channel, dual_channel_changed
        )

        param2ui(
            self.parameters.channel_mixing,
            self.ids.channel_mixing_slider,
            lambda value: value + 128,
        )
        # this is required to update the descriptive labels in the beginning
        self.channel_mixing_changed()

        param2ui(self.parameters.mod_channel, self.ids.mod_channel)
        param2ui(self.parameters.control_channel, self.ids.control_channel)
        param2ui(self.parameters.sweep_channel, self.ids.sweep_channel)
        param2ui(self.parameters.pid_on_slow_enabled, self.ids.slow_control_channel)

        param2ui(self.parameters.polarity_fast_out1, self.ids.polarity_fast_out1)
        param2ui(self.parameters.polarity_fast_out2, self.ids.polarity_fast_out2)
        param2ui(self.parameters.polarity_analog_out0, self.ids.polarity_analog_out0)

        def show_polarity_settings(*args):
            used_channels = set(
                (
                    self.parameters.control_channel.value,
                    self.parameters.sweep_channel.value,
                )
            )

            if self.parameters.pid_on_slow_enabled.value:
                used_channels.add(ANALOG_OUT0)

            self.ids.polarity_selector.setVisible(len(used_channels) > 1)

            def set_visibility(element, channel_id):
                element.setVisible(channel_id in used_channels)

            set_visibility(self.ids.polarity_container_fast_out1, FAST_OUT1)
            set_visibility(self.ids.polarity_container_fast_out2, FAST_OUT2)
            set_visibility(self.ids.polarity_container_analog_out0, ANALOG_OUT0)

        self.parameters.control_channel.on_change(show_polarity_settings)
        self.parameters.sweep_channel.on_change(show_polarity_settings)
        self.parameters.mod_channel.on_change(show_polarity_settings)
        self.parameters.pid_on_slow_enabled.on_change(show_polarity_settings)

        for idx in range(4):
            if idx == 0:
                continue
            name = "analog_out_%d" % idx
            param2ui(
                getattr(self.parameters, name),
                getattr(self.ids, name),
                process_value=lambda v: ANALOG_OUT_V * v,
            )

        def fast_mode_changed(fast_mode_enabled):
            """Disables controls that are irrelevant if fast mode is enabled"""
            widgets_to_disable = (
                self.ids.output_ports_group,
                self.ids.input_ports_group,
            )
            for widget in widgets_to_disable:
                widget.setEnabled(not fast_mode_enabled)

        self.parameters.fast_mode.on_change(fast_mode_changed)

    def channel_mixing_changed(self):
        value = int(self.ids.channel_mixing_slider.value()) - 128
        self.parameters.channel_mixing.value = value
        self.control.write_registers()

        self.update_channel_mixing_slider(value)

    def fast_mode_changed(self):
        self.parameters.fast_mode.value = int(self.ids.fast_mode.checkState() > 0)
        self.control.write_registers()

    def dual_channel_changed(self):
        self.parameters.dual_channel.value = int(self.ids.dual_channel.checkState() > 0)
        self.control.write_registers()

    def update_channel_mixing_slider(self, value):
        a_value, b_value = convert_channel_mixing_value(value)

        self.ids.chain_a_factor.setText("%d" % a_value)
        self.ids.chain_b_factor.setText("%d" % b_value)

    def mod_channel_changed(self, channel):
        self.parameters.mod_channel.value = channel
        self.control.write_registers()

    def control_channel_changed(self, channel):
        self.parameters.control_channel.value = channel
        self.control.write_registers()

    def slow_control_channel_changed(self, channel):
        self.parameters.pid_on_slow_enabled.value = bool(channel)
        self.control.write_registers()

    def sweep_channel_changed(self, channel):
        self.parameters.sweep_channel.value = channel
        self.control.write_registers()

    def polarity_fast_out1_changed(self, polarity):
        self.parameters.polarity_fast_out1.value = bool(polarity)
        self.control.write_registers()

    def polarity_fast_out2_changed(self, polarity):
        self.parameters.polarity_fast_out2.value = bool(polarity)
        self.control.write_registers()

    def polarity_analog_out0_changed(self, polarity):
        self.parameters.polarity_analog_out0.value = bool(polarity)
        self.control.write_registers()



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

import json
import pickle
from os import path

import numpy as np
from linien_common.config import N_COLORS
from linien_gui.utils_gui import color_to_hex, param2ui
from linien_gui.widgets import CustomWidget
from PyQt5 import QtGui, QtWidgets


class ViewPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui("view_panel.ui")

    def ready(self):
        self.ids.export_select_file.clicked.connect(self.do_export_select_file)
        self.ids.export_data.clicked.connect(self.do_export_data)

        self.ids.plot_line_width.setKeyboardTracking(False)
        self.ids.plot_line_width.valueChanged.connect(self.plot_line_width_changed)

        self.ids.plot_line_opacity.setKeyboardTracking(False)
        self.ids.plot_line_opacity.valueChanged.connect(self.plot_line_opacity_changed)

        self.ids.plot_fill_opacity.setKeyboardTracking(False)
        self.ids.plot_fill_opacity.valueChanged.connect(self.plot_fill_opacity_changed)

        for color_idx in range(N_COLORS):
            getattr(self.ids, "edit_color_%d" % color_idx).clicked.connect(
                lambda *args, color_idx=color_idx: self.edit_color(color_idx)
            )

    def edit_color(self, color_idx):
        param = getattr(self.parameters, "plot_color_%d" % color_idx)

        color = QtWidgets.QColorDialog.getColor(QtGui.QColor.fromRgb(*param.value))
        r, g, b, a = color.getRgb()
        print("set color", color_idx, color.getRgb())
        param.value = (r, g, b, a)

    def connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        param2ui(self.parameters.plot_line_width, self.ids.plot_line_width)
        param2ui(self.parameters.plot_line_opacity, self.ids.plot_line_opacity)
        param2ui(self.parameters.plot_fill_opacity, self.ids.plot_fill_opacity)

        def preview_colors(*args):
            for color_idx in range(N_COLORS):
                element = getattr(self.ids, "display_color_%d" % color_idx)
                param = getattr(self.parameters, "plot_color_%d" % color_idx)
                element.setStyleSheet("background-color: " + color_to_hex(param.value))

        for color_idx in range(N_COLORS):
            getattr(self.parameters, "plot_color_%d" % color_idx).on_change(
                preview_colors
            )

    def plot_line_width_changed(self):
        self.parameters.plot_line_width.value = self.ids.plot_line_width.value()

    def plot_line_opacity_changed(self):
        self.parameters.plot_line_opacity.value = self.ids.plot_line_opacity.value()

    def plot_fill_opacity_changed(self):
        self.parameters.plot_fill_opacity.value = self.ids.plot_fill_opacity.value()

    def do_export_select_file(self):
        options = QtWidgets.QFileDialog.Options()
        # options |= QtWidgets.QFileDialog.DontUseNativeDialog
        default_ext = ".json"
        fn, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "QFileDialog.getSaveFileName()",
            "",
            "JSON (*%s)" % default_ext,
            options=options,
        )
        if fn:
            if not fn.endswith(default_ext):
                fn = fn + default_ext
            self.export_fn = fn
            self.ids.export_select_file.setText(
                "File selected: %s" % path.split(fn)[-1]
            )
            self.ids.export_data.setEnabled(True)

    def do_export_data(self):
        fn = self.export_fn
        counter = 0

        while True:
            if counter > 0:
                name, ext = path.splitext(fn)
                fn_with_suffix = name + "-" + str(counter)
                if ext:
                    fn_with_suffix += ext
            else:
                fn_with_suffix = fn

            try:
                open(fn_with_suffix, "r")
                counter += 1
                continue
            except FileNotFoundError:
                break

        print("export data to", fn_with_suffix)

        with open(fn_with_suffix, "w") as f:
            data = pickle.loads(self.parameters.to_plot.value)

            # filter out keys that are not json-able
            for k, v in list(data.items()):
                if isinstance(v, np.ndarray):
                    data[k] = v.tolist()

            json.dump(data, f)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_gui.widgets import CustomWidget
from PyQt5 import QtWidgets


class RightPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        self.parameters.autolock_running.on_change(self.autolock_status_changed)
        self.parameters.optimization_running.on_change(self.optimization_status_changed)
        self.parameters.lock.on_change(self.enable_or_disable_panels)

        def highlight_psd_button(locked):
            self.ids.pid_parameter_optimization_button.setStyleSheet(
                "background: #00aa00;" if locked else ""
            )

        self.parameters.lock.on_change(highlight_psd_button)

    def ready(self):
        self.ids.closeButton.clicked.connect(self.close_app)
        self.ids.shutdownButton.clicked.connect(self.shutdown_server)
        self.ids.openDeviceManagerButton.clicked.connect(self.open_device_manager)
        self.ids.pid_parameter_optimization_button.clicked.connect(self.open_psd_window)

    def close_app(self):
        self.app.close()

    def shutdown_server(self):
        self.app.shutdown()

    def open_psd_window(self):
        self.app.open_psd_window()

    def open_device_manager(self):
        self.app.open_device_manager()

    def autolock_status_changed(self, value):
        if value:
            self.ids.settings_toolbox.setCurrentWidget(self.ids.lockingPanel)

        self.enable_or_disable_panels()

    def optimization_status_changed(self, value):
        if value:
            self.ids.settings_toolbox.setCurrentWidget(self.ids.optimizationPanel)

        self.enable_or_disable_panels()

    def enable_or_disable_panels(self, *args):
        lock = self.parameters.lock.value
        autolock = self.parameters.autolock_running.value
        optimization = self.parameters.optimization_running.value

        def enable_panels(panel_names, condition):
            for panel_name in panel_names:
                getattr(self.ids, panel_name).setEnabled(condition)

        enable_panels(("generalPanel",), not autolock and not optimization and not lock)
        enable_panels(
            ("modSpectroscopyPanel", "viewPanel", "lockingPanel"), not optimization
        )
        enable_panels(("optimizationPanel",), not autolock and not lock)



# Copyright 2018-2022 Benjamin Wiegand <benjamin.wiegand@physik.hu-berlin.de>
# Copyright 2021-2022 Bastian Leykauf <leykauf@physik.hu-berlin.de>
#
# This file is part of Linien and based on redpid.
#
# Linien is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Linien is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linien.  If not, see <http://www.gnu.org/licenses/>.

from linien_common.common import MHz, Vpp
from linien_gui.utils_gui import param2ui
from linien_gui.widgets import CustomWidget
from PyQt5 import QtWidgets


class OptimizationPanel(QtWidgets.QWidget, CustomWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_ui("optimization_panel.ui")

    def ready(self):
        self.ids.start_optimization_btn.clicked.connect(self.start_optimization)
        self.ids.optimization_use_new_parameters.clicked.connect(
            self.use_new_parameters
        )
        self.ids.optimization_abort.clicked.connect(self.abort)
        self.ids.abortOptimizationLineSelection.clicked.connect(self.abort_selection)
        self.ids.abortOptimizationPreparing.clicked.connect(self.abort_preparation)
        self.ids.optimization_channel_selector.currentIndexChanged.connect(
            self.channel_changed
        )
        self.ids.optimization_reset_failed_state.clicked.connect(
            self.reset_failed_state
        )

        for param_name in (
            "optimization_mod_freq_min",
            "optimization_mod_freq_max",
            "optimization_mod_amp_min",
            "optimization_mod_amp_max",
        ):
            element = getattr(self.ids, param_name)
            element.setKeyboardTracking(False)

            def write_parameter(*args, param_name=param_name, element=element):
                getattr(self.parameters, param_name).value = element.value()

            element.valueChanged.connect(write_parameter)

        for param_name in ("optimization_mod_freq", "optimization_mod_amp"):

            def optim_enabled_changed(_, param_name=param_name):
                getattr(self.parameters, param_name + "_enabled").value = int(
                    getattr(self.ids, param_name).checkState()
                )

            getattr(self.ids, param_name).stateChanged.connect(optim_enabled_changed)

    def connection_established(self):
        self.parameters = self.app.parameters
        self.control = self.app.control

        def opt_running_changed(_):
            running = self.parameters.optimization_running.value
            approaching = self.parameters.optimization_approaching.value
            failed = self.parameters.optimization_failed.value

            self.ids.optimization_not_running_container.setVisible(
                not failed and not running
            )
            self.ids.optimization_running_container.setVisible(
                not failed and running and not approaching
            )
            self.ids.optimization_preparing.setVisible(
                not failed and running and approaching
            )
            self.ids.optimization_failed.setVisible(failed)

        self.parameters.optimization_running.on_change(opt_running_changed)
        self.parameters.optimization_approaching.on_change(opt_running_changed)
        self.parameters.optimization_failed.on_change(opt_running_changed)

        def opt_selection_changed(value):
            self.ids.optimization_selecting.setVisible(value)
            self.ids.optimization_not_selecting.setVisible(not value)

        self.parameters.optimization_selection.on_change(opt_selection_changed)

        def mod_param_changed(_):
            dual_channel = self.parameters.dual_channel.value
            channel = self.parameters.optimization_channel.value
            optimized = self.parameters.optimization_optimized_parameters.value

            self.ids.optimization_display_parameters.setText(
                (
                    "<br />\n"
                    "<b>current parameters</b>: "
                    " %.2f&nbsp;MHz, %.2f&nbsp;Vpp, %.2f&nbsp;deg<br />\n"
                    "<b>optimized parameters</b>: "
                    "%.2f&nbsp;MHz, %.2f&nbsp;Vpp, %.2f&nbsp;deg\n"
                    "<br />"
                )
                % (
                    self.parameters.modulation_frequency.value / MHz,
                    self.parameters.modulation_amplitude.value / Vpp,
                    (
                        self.parameters.demodulation_phase_a,
                        self.parameters.demodulation_phase_b,
                    )[0 if not dual_channel else (0, 1)[channel]].value,
                    optimized[0] / MHz,
                    optimized[1] / Vpp,
                    optimized[2],
                )
            )

        for p in (
            self.parameters.modulation_amplitude,
            self.parameters.modulation_frequency,
            self.parameters.demodulation_phase_a,
        ):
            p.on_change(mod_param_changed)

        def improvement_changed(improvement):
            self.ids.optimization_improvement.setText("%d %%" % (improvement * 100))

        self.parameters.optimization_improvement.on_change(improvement_changed)

        param2ui(
            self.parameters.optimization_mod_freq_enabled,
            self.ids.optimization_mod_freq,
        )
        param2ui(
            self.parameters.optimization_mod_freq_min,
            self.ids.optimization_mod_freq_min,
        )
        param2ui(
            self.parameters.optimization_mod_freq_max,
            self.ids.optimization_mod_freq_max,
        )
        param2ui(
            self.parameters.optimization_mod_amp_enabled, self.ids.optimization_mod_amp
        )
        param2ui(
            self.parameters.optimization_mod_amp_min, self.ids.optimization_mod_amp_min
        )
        param2ui(
            self.parameters.optimization_mod_amp_max, self.ids.optimization_mod_amp_max
        )

        def dual_channel_changed(value):
            self.ids.optimization_channel_selector_box.setVisible(value)

        self.parameters.dual_channel.on_change(dual_channel_changed)

        def fast_mode_changed(fast_mode_enabled):
            """Disable this panel if fast mode is enabled (nothing to optimize)."""
            self.setEnabled(not fast_mode_enabled)

        self.parameters.fast_mode.on_change(fast_mode_changed)

        def fast_mode_changed(fast_mode_enabled):
            """Disable this panel if fast mode is enabled (nothing to optimize)."""
            self.setEnabled(not fast_mode_enabled)

        self.parameters.fast_mode.on_change(fast_mode_changed)

    def start_optimization(self):
        self.parameters.optimization_selection.value = True

    def abort_selection(self):
        self.parameters.optimization_selection.value = False

    def abort_preparation(self):
        self.parameters.task.value.stop(False)

    def abort(self):
        self.parameters.task.value.stop(False)

    def use_new_parameters(self):
        self.parameters.task.value.stop(True)

    def channel_changed(self, channel):
        self.parameters.optimization_channel.value = channel

    def reset_failed_state(self):
        self.parameters.optimization_failed.value = False
